@@ -1,19 +1,19 @@
-use std::cell::RefCell;
-use std::collections::HashMap;
-use std::fmt;
-use std::ops::Index;
-use std::sync::Arc;
-
-use log::debug;
-use pcre2_sys::{
-    PCRE2_CASELESS, PCRE2_DOTALL, PCRE2_EXTENDED, PCRE2_MULTILINE,
-    PCRE2_UCP, PCRE2_UTF, PCRE2_NO_UTF_CHECK, PCRE2_UNSET,
-    PCRE2_NEWLINE_ANYCRLF,
+use std::{
+    collections::HashMap,
+    panic::{RefUnwindSafe, UnwindSafe},
+    sync::Arc,
 };
-use thread_local::CachedThreadLocal;
 
-use crate::error::Error;
-use crate::ffi::{Code, CompileContext, MatchConfig, MatchData};
+use pcre2_sys::{
+    PCRE2_CASELESS, PCRE2_DOTALL, PCRE2_EXTENDED, PCRE2_MATCH_INVALID_UTF,
+    PCRE2_MULTILINE, PCRE2_NEWLINE_ANYCRLF, PCRE2_UCP, PCRE2_UNSET, PCRE2_UTF,
+};
+
+use crate::{
+    error::Error,
+    ffi::{Code, CompileContext, MatchConfig, MatchData},
+    pool::{Pool, PoolGuard},
+};
 
 /// Match represents a single match of a regex in a subject string.
 ///
@@ -72,8 +72,6 @@ struct Config {
     ucp: bool,
     /// PCRE2_UTF
     utf: bool,
-    /// PCRE2_NO_UTF_CHECK
-    utf_check: bool,
     /// use pcre2_jit_compile
     jit: JITChoice,
     /// Match-time specific configuration knobs.
@@ -100,7 +98,6 @@ impl Default for Config {
             crlf: false,
             ucp: false,
             utf: false,
-            utf_check: true,
             jit: JITChoice::Never,
             match_config: MatchConfig::default(),
         }
@@ -141,6 +138,7 @@ impl RegexBuilder {
         if self.config.ucp {
             options |= PCRE2_UCP;
             options |= PCRE2_UTF;
+            options |= PCRE2_MATCH_INVALID_UTF;
         }
         if self.config.utf {
             options |= PCRE2_UTF;
@@ -160,7 +158,7 @@ impl RegexBuilder {
             }
             JITChoice::Attempt => {
                 if let Err(err) = code.jit_compile() {
-                    debug!("JIT compilation failed: {}", err);
+                    log::debug!("JIT compilation failed: {}", err);
                 }
             }
         }
@@ -171,13 +169,21 @@ impl RegexBuilder {
                 idx.insert(name.to_string(), i);
             }
         }
+        let code = Arc::new(code);
+        let match_data = {
+            let config = self.config.match_config.clone();
+            let code = Arc::clone(&code);
+            let create: MatchDataPoolFn =
+                Box::new(move || MatchData::new(config.clone(), &code));
+            Pool::new(create)
+        };
         Ok(Regex {
             config: Arc::new(self.config.clone()),
             pattern: pattern.to_string(),
-            code: Arc::new(code),
+            code,
             capture_names: Arc::new(capture_names),
             capture_names_idx: Arc::new(idx),
-            match_data: CachedThreadLocal::new(),
+            match_data,
         })
     }
 
@@ -260,35 +266,28 @@ impl RegexBuilder {
     /// as when this is disabled, `.` will any single byte (except for `\n` in
     /// both cases, unless "dot all" mode is enabled).
     ///
-    /// Note that when UTF matching mode is enabled, every search performed
-    /// will do a UTF-8 validation check, which can impact performance. The
-    /// UTF-8 check can be disabled via the `disable_utf_check` option, but it
-    /// is undefined behavior to enable UTF matching mode and search invalid
-    /// UTF-8.
-    ///
     /// This is disabled by default.
     pub fn utf(&mut self, yes: bool) -> &mut RegexBuilder {
         self.config.utf = yes;
         self
     }
 
-    /// When UTF matching mode is enabled, this will disable the UTF checking
-    /// that PCRE2 will normally perform automatically. If UTF matching mode
-    /// is not enabled, then this has no effect.
+    /// This is now deprecated and is a no-op.
     ///
-    /// UTF checking is enabled by default when UTF matching mode is enabled.
-    /// If UTF matching mode is enabled and UTF checking is enabled, then PCRE2
-    /// will return an error if you attempt to search a subject string that is
-    /// not valid UTF-8.
+    /// Previously, this option permitted disabling PCRE2's UTF-8 validity
+    /// check, which could result in undefined behavior if the haystack was
+    /// not valid UTF-8. But PCRE2 introduced a new option, `PCRE2_MATCH_INVALID_UTF`,
+    /// in 10.34 which this crate always sets. When this option is enabled,
+    /// PCRE2 claims to not have undefined behavior when the haystack is
+    /// invalid UTF-8.
     ///
-    /// # Safety
-    ///
-    /// It is undefined behavior to disable the UTF check in UTF matching mode
-    /// and search a subject string that is not valid UTF-8. When the UTF check
-    /// is disabled, callers must guarantee that the subject string is valid
-    /// UTF-8.
-    pub unsafe fn disable_utf_check(&mut self) -> &mut RegexBuilder {
-        self.config.utf_check = false;
+    /// Therefore, disabling the UTF-8 check is not something that is exposed
+    /// by this crate.
+    #[deprecated(
+        since = "0.2.4",
+        note = "now a no-op due to new PCRE2 features"
+    )]
+    pub fn disable_utf_check(&mut self) -> &mut RegexBuilder {
         self
     }
 
@@ -366,30 +365,32 @@ pub struct Regex {
     capture_names: Arc<Vec<Option<String>>>,
     /// A map from capture group name to capture group index.
     capture_names_idx: Arc<HashMap<String, usize>>,
-    /// Mutable scratch data used by PCRE2 during matching.
-    ///
-    /// We use the same strategy as Rust's regex crate here, such that each
-    /// thread gets its own match data to support using a Regex object from
-    /// multiple threads simultaneously. If some match data doesn't exist for
-    /// a thread, then a new one is created on demand.
-    match_data: CachedThreadLocal<RefCell<MatchData>>,
+    /// A pool of mutable scratch data used by PCRE2 during matching.
+    match_data: MatchDataPool,
 }
 
 impl Clone for Regex {
     fn clone(&self) -> Regex {
+        let match_data = {
+            let config = self.config.match_config.clone();
+            let code = Arc::clone(&self.code);
+            let create: MatchDataPoolFn =
+                Box::new(move || MatchData::new(config.clone(), &code));
+            Pool::new(create)
+        };
         Regex {
             config: Arc::clone(&self.config),
             pattern: self.pattern.clone(),
             code: Arc::clone(&self.code),
             capture_names: Arc::clone(&self.capture_names),
             capture_names_idx: Arc::clone(&self.capture_names_idx),
-            match_data: CachedThreadLocal::new(),
+            match_data,
         }
     }
 }
 
-impl fmt::Debug for Regex {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+impl std::fmt::Debug for Regex {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "Regex({:?})", self.pattern)
     }
 }
@@ -475,7 +476,7 @@ impl Regex {
         Matches {
             re: self,
             match_data: self.match_data(),
-            subject: subject,
+            subject,
             last_end: 0,
             last_match: None,
         }
@@ -546,8 +547,8 @@ impl Regex {
     ) -> Result<Option<Captures<'s>>, Error> {
         let mut locs = self.capture_locations();
         Ok(self.captures_read(&mut locs, subject)?.map(move |_| Captures {
-            subject: subject,
-            locs: locs,
+            subject,
+            locs,
             idx: Arc::clone(&self.capture_names_idx),
         }))
     }
@@ -585,12 +586,7 @@ impl Regex {
         &'r self,
         subject: &'s [u8],
     ) -> CaptureMatches<'r, 's> {
-        CaptureMatches {
-            re: self,
-            subject: subject,
-            last_end: 0,
-            last_match: None,
-        }
+        CaptureMatches { re: self, subject, last_end: 0, last_match: None }
     }
 }
 
@@ -614,18 +610,13 @@ impl Regex {
             subject.len()
         );
 
-        let mut options = 0;
-        if !self.config.utf_check {
-            options |= PCRE2_NO_UTF_CHECK;
-        }
-
-        let match_data = self.match_data();
-        let mut match_data = match_data.borrow_mut();
-        // SAFETY: The only unsafe PCRE2 option we potentially use here is
-        // PCRE2_NO_UTF_CHECK, and that only occurs if the caller executes the
-        // `disable_utf_check` method, which propagates the safety contract to
-        // the caller.
-        Ok(unsafe { match_data.find(&self.code, subject, start, options)? })
+        let options = 0;
+        let mut match_data = self.match_data();
+        // SAFETY: We don't use any dangerous PCRE2 options.
+        let res =
+            unsafe { match_data.find(&self.code, subject, start, options) };
+        PoolGuard::put(match_data);
+        res
     }
 
     /// Returns the same as find, but starts the search at the given
@@ -639,7 +630,11 @@ impl Regex {
         subject: &'s [u8],
         start: usize,
     ) -> Result<Option<Match<'s>>, Error> {
-        self.find_at_with_match_data(self.match_data(), subject, start)
+        let mut match_data = self.match_data();
+        let res =
+            self.find_at_with_match_data(&mut match_data, subject, start);
+        PoolGuard::put(match_data);
+        res
     }
 
     /// Like find_at, but accepts match data instead of acquiring one itself.
@@ -649,7 +644,7 @@ impl Regex {
     #[inline(always)]
     fn find_at_with_match_data<'s>(
         &self,
-        match_data: &RefCell<MatchData>,
+        match_data: &mut MatchDataPoolGuard<'_>,
         subject: &'s [u8],
         start: usize,
     ) -> Result<Option<Match<'s>>, Error> {
@@ -660,22 +655,14 @@ impl Regex {
             subject.len()
         );
 
-        let mut options = 0;
-        if !self.config.utf_check {
-            options |= PCRE2_NO_UTF_CHECK;
-        }
-
-        let mut match_data = match_data.borrow_mut();
-        // SAFETY: The only unsafe PCRE2 option we potentially use here is
-        // PCRE2_NO_UTF_CHECK, and that only occurs if the caller executes the
-        // `disable_utf_check` method, which propagates the safety contract to
-        // the caller.
+        let options = 0;
+        // SAFETY: We don't use any dangerous PCRE2 options.
         if unsafe { !match_data.find(&self.code, subject, start, options)? } {
             return Ok(None);
         }
         let ovector = match_data.ovector();
         let (s, e) = (ovector[0], ovector[1]);
-        Ok(Some(Match::new(&subject[s..e], s, e)))
+        Ok(Some(Match::new(&subject, s, e)))
     }
 
     /// This is like `captures`, but uses
@@ -715,20 +702,14 @@ impl Regex {
             subject.len()
         );
 
-        let mut options = 0;
-        if !self.config.utf_check {
-            options |= PCRE2_NO_UTF_CHECK;
-        }
-        // SAFETY: The only unsafe PCRE2 option we potentially use here is
-        // PCRE2_NO_UTF_CHECK, and that only occurs if the caller executes the
-        // `disable_utf_check` method, which propagates the safety contract to
-        // the caller.
+        let options = 0;
+        // SAFETY: We don't use any dangerous PCRE2 options.
         if unsafe { !locs.data.find(&self.code, subject, start, options)? } {
             return Ok(None);
         }
         let ovector = locs.data.ovector();
         let (s, e) = (ovector[0], ovector[1]);
-        Ok(Some(Match::new(&subject[s..e], s, e)))
+        Ok(Some(Match::new(&subject, s, e)))
     }
 }
 
@@ -771,9 +752,8 @@ impl Regex {
         }
     }
 
-    fn match_data(&self) -> &RefCell<MatchData> {
-        let create = || RefCell::new(self.new_match_data());
-        self.match_data.get_or(create)
+    fn match_data(&self) -> MatchDataPoolGuard<'_> {
+        self.match_data.get()
     }
 
     fn new_match_data(&self) -> MatchData {
@@ -805,8 +785,8 @@ impl Clone for CaptureLocations {
     }
 }
 
-impl fmt::Debug for CaptureLocations {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+impl std::fmt::Debug for CaptureLocations {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         let mut offsets: Vec<Option<usize>> = vec![];
         for &offset in self.data.ovector() {
             if offset == PCRE2_UNSET {
@@ -915,16 +895,16 @@ impl<'s> Captures<'s> {
     }
 }
 
-impl<'s> fmt::Debug for Captures<'s> {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+impl<'s> std::fmt::Debug for Captures<'s> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         f.debug_tuple("Captures").field(&CapturesDebug(self)).finish()
     }
 }
 
 struct CapturesDebug<'c, 's: 'c>(&'c Captures<'s>);
 
-impl<'c, 's> fmt::Debug for CapturesDebug<'c, 's> {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+impl<'c, 's> std::fmt::Debug for CapturesDebug<'c, 's> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         fn escape_bytes(bytes: &[u8]) -> String {
             let mut s = String::new();
             for &b in bytes {
@@ -946,9 +926,11 @@ impl<'c, 's> fmt::Debug for CapturesDebug<'c, 's> {
             self.0.idx.iter().map(|(a, b)| (b, a)).collect();
         let mut map = f.debug_map();
         for slot in 0..self.0.len() {
-            let m = self.0.locs.get(slot).map(|(s, e)| {
-                escape_bytes(&self.0.subject[s..e])
-            });
+            let m = self
+                .0
+                .locs
+                .get(slot)
+                .map(|(s, e)| escape_bytes(&self.0.subject[s..e]));
             if let Some(name) = slot_to_name.get(&slot) {
                 map.entry(&name, &m);
             } else {
@@ -970,11 +952,12 @@ impl<'c, 's> fmt::Debug for CapturesDebug<'c, 's> {
 /// # Panics
 ///
 /// If there is no group at the given index.
-impl<'s> Index<usize> for Captures<'s> {
+impl<'s> std::ops::Index<usize> for Captures<'s> {
     type Output = [u8];
 
     fn index(&self, i: usize) -> &[u8] {
-        self.get(i).map(|m| m.as_bytes())
+        self.get(i)
+            .map(|m| m.as_bytes())
             .unwrap_or_else(|| panic!("no group at index '{}'", i))
     }
 }
@@ -991,11 +974,12 @@ impl<'s> Index<usize> for Captures<'s> {
 /// # Panics
 ///
 /// If there is no group named by the given value.
-impl<'s, 'i> Index<&'i str> for Captures<'s> {
+impl<'s, 'i> std::ops::Index<&'i str> for Captures<'s> {
     type Output = [u8];
 
     fn index<'a>(&'a self, name: &'i str) -> &'a [u8] {
-        self.name(name).map(|m| m.as_bytes())
+        self.name(name)
+            .map(|m| m.as_bytes())
             .unwrap_or_else(|| panic!("no group named '{}'", name))
     }
 }
@@ -1011,7 +995,7 @@ impl<'s, 'i> Index<&'i str> for Captures<'s> {
 /// lifetime of the subject string.
 pub struct Matches<'r, 's> {
     re: &'r Regex,
-    match_data: &'r RefCell<MatchData>,
+    match_data: MatchDataPoolGuard<'r>,
     subject: &'s [u8],
     last_end: usize,
     last_match: Option<usize>,
@@ -1025,7 +1009,7 @@ impl<'r, 's> Iterator for Matches<'r, 's> {
             return None;
         }
         let res = self.re.find_at_with_match_data(
-            self.match_data,
+            &mut self.match_data,
             self.subject,
             self.last_end,
         );
@@ -1074,11 +1058,8 @@ impl<'r, 's> Iterator for CaptureMatches<'r, 's> {
             return None;
         }
         let mut locs = self.re.capture_locations();
-        let res = self.re.captures_read_at(
-            &mut locs,
-            self.subject,
-            self.last_end,
-        );
+        let res =
+            self.re.captures_read_at(&mut locs, self.subject, self.last_end);
         let m = match res {
             Err(err) => return Some(Err(err)),
             Ok(None) => return None,
@@ -1100,12 +1081,24 @@ impl<'r, 's> Iterator for CaptureMatches<'r, 's> {
         self.last_match = Some(m.end());
         Some(Ok(Captures {
             subject: self.subject,
-            locs: locs,
+            locs,
             idx: Arc::clone(&self.re.capture_names_idx),
         }))
     }
 }
 
+/// A type alias for our pool of `MatchData` that fixes the type parameters to
+/// what we actually use in practice.
+type MatchDataPool = Pool<MatchData, MatchDataPoolFn>;
+
+/// Same as above, but for the guard returned by a pool.
+type MatchDataPoolGuard<'a> = PoolGuard<'a, MatchData, MatchDataPoolFn>;
+
+/// The type of the closure we use to create new caches. We need to spell out
+/// all of the marker traits or else we risk leaking !MARKER impls.
+type MatchDataPoolFn =
+    Box<dyn Fn() -> MatchData + Send + Sync + UnwindSafe + RefUnwindSafe>;
+
 #[cfg(test)]
 mod tests {
     use super::{Regex, RegexBuilder};
@@ -1136,114 +1129,72 @@ mod tests {
 
     #[test]
     fn caseless() {
-        let re = RegexBuilder::new()
-            .caseless(true)
-            .build("a")
-            .unwrap();
+        let re = RegexBuilder::new().caseless(true).build("a").unwrap();
         assert!(re.is_match(b("A")).unwrap());
 
-        let re = RegexBuilder::new()
-            .caseless(true)
-            .ucp(true)
-            .build("β")
-            .unwrap();
+        let re =
+            RegexBuilder::new().caseless(true).ucp(true).build("β").unwrap();
         assert!(re.is_match(b("Β")).unwrap());
     }
 
     #[test]
     fn crlf() {
-        let re = RegexBuilder::new()
-            .crlf(true)
-            .build("a$")
-            .unwrap();
+        let re = RegexBuilder::new().crlf(true).build("a$").unwrap();
         let m = re.find(b("a\r\n")).unwrap().unwrap();
         assert_eq!(m.as_pair(), (0, 1));
     }
 
     #[test]
     fn dotall() {
-        let re = RegexBuilder::new()
-            .dotall(false)
-            .build(".")
-            .unwrap();
+        let re = RegexBuilder::new().dotall(false).build(".").unwrap();
         assert!(!re.is_match(b("\n")).unwrap());
 
-        let re = RegexBuilder::new()
-            .dotall(true)
-            .build(".")
-            .unwrap();
+        let re = RegexBuilder::new().dotall(true).build(".").unwrap();
         assert!(re.is_match(b("\n")).unwrap());
     }
 
     #[test]
     fn extended() {
-        let re = RegexBuilder::new()
-            .extended(true)
-            .build("a b c")
-            .unwrap();
+        let re = RegexBuilder::new().extended(true).build("a b c").unwrap();
         assert!(re.is_match(b("abc")).unwrap());
     }
 
     #[test]
     fn multi_line() {
-        let re = RegexBuilder::new()
-            .multi_line(false)
-            .build("^abc$")
-            .unwrap();
+        let re = RegexBuilder::new().multi_line(false).build("^abc$").unwrap();
         assert!(!re.is_match(b("foo\nabc\nbar")).unwrap());
 
-        let re = RegexBuilder::new()
-            .multi_line(true)
-            .build("^abc$")
-            .unwrap();
+        let re = RegexBuilder::new().multi_line(true).build("^abc$").unwrap();
         assert!(re.is_match(b("foo\nabc\nbar")).unwrap());
     }
 
     #[test]
     fn ucp() {
-        let re = RegexBuilder::new()
-            .ucp(false)
-            .build(r"\w")
-            .unwrap();
+        let re = RegexBuilder::new().ucp(false).build(r"\w").unwrap();
         assert!(!re.is_match(b("β")).unwrap());
 
-        let re = RegexBuilder::new()
-            .ucp(true)
-            .build(r"\w")
-            .unwrap();
+        let re = RegexBuilder::new().ucp(true).build(r"\w").unwrap();
         assert!(re.is_match(b("β")).unwrap());
     }
 
     #[test]
     fn utf() {
-        let re = RegexBuilder::new()
-            .utf(false)
-            .build(".")
-            .unwrap();
+        let re = RegexBuilder::new().utf(false).build(".").unwrap();
         assert_eq!(re.find(b("β")).unwrap().unwrap().as_pair(), (0, 1));
 
-        let re = RegexBuilder::new()
-            .utf(true)
-            .build(".")
-            .unwrap();
+        let re = RegexBuilder::new().utf(true).build(".").unwrap();
         assert_eq!(re.find(b("β")).unwrap().unwrap().as_pair(), (0, 2));
     }
 
     #[test]
     fn jit4lyfe() {
         if is_jit_available() {
-            let re = RegexBuilder::new()
-                .jit(true)
-                .build(r"\w")
-                .unwrap();
+            let re = RegexBuilder::new().jit(true).build(r"\w").unwrap();
             assert!(re.is_match(b("a")).unwrap());
         } else {
             // Check that if JIT isn't enabled, then we get an error if we
             // require JIT.
-            RegexBuilder::new()
-                .jit(true)
-                .build(r"\w")
-                .unwrap_err();
+            RegexBuilder::new().jit(true).build(r"\w").unwrap_err();
         }
     }
 
@@ -1252,10 +1203,8 @@ mod tests {
     // If the JIT isn't available, then in this test, we simply don't use it.
     #[test]
     fn jit_if_available() {
-        let re = RegexBuilder::new()
-            .jit_if_available(true)
-            .build(r"\w")
-            .unwrap();
+        let re =
+            RegexBuilder::new().jit_if_available(true).build(r"\w").unwrap();
         assert!(re.is_match(b("a")).unwrap());
     }
 
@@ -1274,32 +1223,28 @@ mod tests {
 
     #[test]
     fn utf_with_invalid_data() {
-        let re = RegexBuilder::new()
-            .build(r".")
-            .unwrap();
+        let re = RegexBuilder::new().build(r".").unwrap();
         assert_eq!(re.find(b"\xFF").unwrap().unwrap().as_pair(), (0, 1));
 
-        let re = RegexBuilder::new()
-            .utf(true)
-            .build(r".")
-            .unwrap();
+        let re = RegexBuilder::new().utf(true).build(r".").unwrap();
         assert!(re.find(b"\xFF").is_err());
     }
 
     #[test]
     fn capture_names() {
         let re = RegexBuilder::new()
-            .build(
-                r"(?P<foo>abc)|(def)|(?P<a>ghi)|(?P<springsteen>jkl)"
-            )
+            .build(r"(?P<foo>abc)|(def)|(?P<a>ghi)|(?P<springsteen>jkl)")
             .unwrap();
-        assert_eq!(re.capture_names().to_vec(), vec![
-            None,
-            Some("foo".to_string()),
-            None,
-            Some("a".to_string()),
-            Some("springsteen".to_string()),
-        ]);
+        assert_eq!(
+            re.capture_names().to_vec(),
+            vec![
+                None,
+                Some("foo".to_string()),
+                None,
+                Some("a".to_string()),
+                Some("springsteen".to_string()),
+            ]
+        );
 
         // Test our internal map as well.
         assert_eq!(re.capture_names_idx.len(), 3);
@@ -1326,9 +1271,10 @@ mod tests {
         assert_eq!(find_iter_tuples(&re, b"\n"), vec![(0, 0)]);
         assert_eq!(find_iter_tuples(&re, b"\n\n"), vec![(0, 0), (1, 1)]);
         assert_eq!(find_iter_tuples(&re, b"\na\n"), vec![(0, 0), (1, 1)]);
-        assert_eq!(find_iter_tuples(&re, b"\na\n\n"), vec![
-            (0, 0), (1, 1), (3, 3),
-        ]);
+        assert_eq!(
+            find_iter_tuples(&re, b"\na\n\n"),
+            vec![(0, 0), (1, 1), (3, 3),]
+        );
     }
 
     #[test]
@@ -1338,9 +1284,10 @@ mod tests {
         assert_eq!(cap_iter_tuples(&re, b"\n"), vec![(0, 0)]);
         assert_eq!(cap_iter_tuples(&re, b"\n\n"), vec![(0, 0), (1, 1)]);
         assert_eq!(cap_iter_tuples(&re, b"\na\n"), vec![(0, 0), (1, 1)]);
-        assert_eq!(cap_iter_tuples(&re, b"\na\n\n"), vec![
-            (0, 0), (1, 1), (3, 3),
-        ]);
+        assert_eq!(
+            cap_iter_tuples(&re, b"\na\n\n"),
+            vec![(0, 0), (1, 1), (3, 3),]
+        );
     }
 
     #[test]
@@ -1377,9 +1324,50 @@ mod tests {
         let re = RegexBuilder::new()
             .ucp(true)
             .jit(true)
-            .max_jit_stack_size(Some(1<<20))
+            .max_jit_stack_size(Some(1 << 20))
             .build(r"((((\w{10})){100}))+")
             .unwrap();
         assert!(re.is_match(hay.as_bytes()).unwrap());
     }
+
+    #[test]
+    fn find_start_end_and_as_bytes() {
+        let hay =
+            "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+        let pattern = r"
+            (?x)    (?#: Allow comments and whitespace.)
+
+            [a-z]   (?#: Lowercase letter.)
+            +       (?#: One or more times.)
+            ";
+        let re = RegexBuilder::new()
+            .extended(true)
+            .utf(true)
+            .jit(true)
+            .build(pattern)
+            .unwrap();
+        let matched = re.find(hay.as_bytes()).unwrap().unwrap();
+        assert_eq!(matched.start(), 10);
+        assert_eq!(matched.end(), 10 + 26);
+        assert_eq!(matched.as_bytes(), b"abcdefghijklmnopqrstuvwxyz");
+    }
+
+    #[test]
+    fn find_utf_emoji_as_bytes() {
+        let hay = "0123456789😀👍🏼🎉abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+        let pattern = r"(*UTF)
+            (?x)    (?#: Allow comments and whitespace.)
+
+            [^\N{U+0000}-\N{U+007F}]    (?#: Non-ascii code points.)
+            +                           (?#: One or more times.)
+            ";
+        let re = RegexBuilder::new()
+            .extended(true)
+            .utf(true)
+            .jit(true)
+            .build(pattern)
+            .unwrap();
+        let matched = re.find(hay.as_bytes()).unwrap().unwrap();
+        assert_eq!(matched.as_bytes(), "😀👍🏼🎉".as_bytes());
+    }
 }
